#pragma once

#include <vector>
#include <tuple>
#include <string>
#include <sstream>
#include <iomanip>
#include <cstring>
#include <chrono>
#include <cfitsio/fitsio.h>
#include <regex>
#include <filesystem>

#include <constants.h>
#include <Astri_MA_Generic.h>
#include <sofa.h>

namespace FitsElements {
namespace HK101{

static int num_pdm = 37;
static int num_pdm_pixel_rate = 64;


static std::string time_utc;
double start_time;
double stop_time;

namespace last_event_time {
	int year, month, day, hours, minutes,seconds;
};

namespace filename_params {
	std::string date, tel_id, obs_mode, run_id, mode_id, daq_id, target_id;
};

namespace database_params {
	std::string orig_id, prog_id, sbl_id;
	uint16_t ntel;
	float ra_obj, dec_obj, ra_pnt, dec_pnt;
};

// primary HK101HDU {keyword, value, comment}
static std::vector<std::tuple<std::string, std::string, std::string>> primary_hdu;
// HK101 hdu events {type, format, unit, format_CODE}
static std::vector<std::tuple<std::string, std::string, std::string, int >> events_hdu;
static char **ttype; // init in FitsFunctions::initFitsSchema
static char **tunit;// init in FitsFunctions::initFitsSchema
static char **tform; // init in FitsFunctions::initFitsSchema
static int ncol; // init in FitsFunctions::initFitsSchema
static const char* extname = "EVENTS";

}};

namespace FitsFunctions {

static void initHK101FitsSchema(){

	// HK101 hdu events {type, format, unit, format_CODE}
	FitsElements::HK101::events_hdu = {
		{"VT","1L","", TLOGICAL},
		{"TIME_UTC","19A","YYYY-MM-DDThh:mm:ss", TSTRING},
		{"TIME_S","1V","s", TUINT},
		{"TIME_NS","1V","ns", TUINT},
		{"SSC","1U","", TUSHORT},
		{"PKT_LEN","1U","", TUSHORT},
		{"EVT_CTR","1V","", TUINT},
		{"LID","1L","", TLOGICAL},
		{"FIB_ST","1L","", TLOGICAL},
		{"FIB_CONT","1L","", TLOGICAL},
		{"FIB_PULS","1L","", TLOGICAL},
		{"RGB_CONT","1I","", TSHORT},
		{"RGB_PULS","1I","", TSHORT},
		{"CAM_TMP1","1I","(°C + 40)*100", TSHORT},
		{"CAM_TMP2","1I","(°C + 40)*100", TSHORT},
		{"CAM_TMP3","1I","(°C + 40)*100", TSHORT},
		{"CAM_TMP4","1I","(°C + 40)*100", TSHORT},
		{"CAM_HUM1","1I","% * 100", TSHORT},
		{"CAM_HUM2","1I","% * 100", TSHORT},
		{"CAM_HUM3","1I","% * 100", TSHORT},
		{"CAM_HUM4","1I","% * 100", TSHORT},
		{"BL_FRQ","1I","Hz", TSHORT},
		{"BL_DUR","1I","ns", TSHORT},
		{"BL_V","1I","V", TSHORT},
		{"GL_FRQ","1I","Hz", TSHORT},
		{"GL_DUR","1I","ns", TSHORT},
		{"GL_V","1I","V", TSHORT},
		{"LD_TMP","1U","", TUSHORT},
		{"CAM_RTM","1I","", TSHORT},
		{"TCAM_RTM","1I","", TSHORT},
		{"DEAD_TIME","1I","%value * 100", TSHORT},
		{"PDM_VAL","37L","", TLOGICAL},
		{"PDM_TMP1","37U","", TUSHORT},
		{"PDM_TMP2","37U","", TUSHORT},
		{"PDM_TMP3","37U","", TUSHORT},
		{"PDM_CBT","37U","", TUSHORT},
		{"PDM_CBV1","37U","", TUSHORT},
		{"PDM_CBV2","37U","", TUSHORT},
		{"PDM_CBC1","37U","", TUSHORT},
		{"PDM_CBC2","37U","", TUSHORT},
		{"PDM_CBC3","37U","", TUSHORT},
		{"FPGA_TMP","37U","", TUSHORT},
		{"FPGA_V","37U","", TUSHORT},
		{"SIPM_HV","37U","", TUSHORT},
		{"SIPM_C","37V","", TUINT},
		{"CAM_RT","37U","", TUSHORT},
		{"PDM_TRT","37U","", TUSHORT},
		};

	// PDM Pixel rate
	for(int i = 0; i < FitsElements::HK101::num_pdm; i++){
		std::stringstream fieldName;
		fieldName << "PDM";
		fieldName << std::setfill('0') << std::setw(2) << i; // 01, 02...37
		fieldName << "PRT";
		FitsElements::HK101::events_hdu.push_back({fieldName.str(),"64V","", TUINT});
	}

	FitsElements::HK101::ncol = FitsElements::HK101::events_hdu.size();
	FitsElements::HK101::ttype = new char*[FitsElements::HK101::ncol];
	FitsElements::HK101::tform = new char*[FitsElements::HK101::ncol];
	FitsElements::HK101::tunit = new char*[FitsElements::HK101::ncol];
// init char** of type, form, uinit
	for (int i = 0; i < FitsElements::HK101::ncol; i++) {
		std::string type = std::get<0>(FitsElements::HK101::events_hdu[i]);
		std::string form = std::get<1>(FitsElements::HK101::events_hdu[i]);
		std::string unit = std::get<2>(FitsElements::HK101::events_hdu[i]);
		FitsElements::HK101::ttype[i] = new char[type.size() + 1];
		FitsElements::HK101::tform[i] = new char[form.size() + 1];
		FitsElements::HK101::tunit[i] = new char[unit.size() + 1];
		std::strcpy(FitsElements::HK101::ttype[i], type.c_str());
		std::strcpy(FitsElements::HK101::tform[i], form.c_str());
		std::strcpy(FitsElements::HK101::tunit[i], unit.c_str());
	}
}

static int initFitsFileHK101(fitsfile *fits_file, std::string fitsFile, std::string rawFile) { // rawFileName to extract params
	using namespace FitsElements::HK101;
	std::regex pattern(RAW_FILE_NAMING_CONV);
	std::smatch match;
	std::filesystem::path rawFilePath(rawFile);
	std::string rawFileName(rawFilePath.filename());
	std::filesystem::path fitsFilePath(fitsFile);
	std::string fitsFileName(fitsFilePath.filename());
	if (!std::regex_match(rawFileName, match, pattern)) {
		std::cerr << "Error extracting params from file name: " << rawFileName << std::endl;
		return -1;
	}
	filename_params::date = match[1];
	filename_params::tel_id = match[2];
	filename_params::target_id = match[3];
	filename_params::obs_mode = match[4];
	filename_params::run_id = match[5];
	filename_params::mode_id = match[6];
	filename_params::daq_id = match[7];

	// TODO check db connection and extract database_params

	int status = 0;

	int simple = 1;
	fits_write_key(fits_file, TLOGICAL, "SIMPLE", &simple, "File FITS di esempio", &status);
	int bitpix = 16;
	fits_write_key(fits_file, TINT, "BITPIX", &bitpix, "Numero di bit per pixel", &status);
	int naxis = 0;
	fits_write_key(fits_file, TINT, "NAXIS", &naxis, "Numero di dimensioni", &status);
	int extend = 1;
	fits_write_key(fits_file, TLOGICAL, "EXTEND", &extend, "Fits Dataset may contain extensions", &status);
	fits_write_key_str(fits_file, "COMMENT", "FITS (Flexible Image Transport System) format is defined in 'Astronomy and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H", "comment", &status);
	fits_write_key_str(fits_file, "TELESCOP", "ASTRI-MA","Name of telescope generating data", &status);
	fits_write_key_str(fits_file, "INSTRUME","CAMERA","Name of instrument generating data", &status);
	fits_write_key_str(fits_file, "TEL_ID", filename_params::tel_id.c_str(),"Telescope ID", &status);
	fits_write_key_str(fits_file, "ORIGIN","ASTRIDPS","Name of organization making file", &status);
	fits_write_key_str(fits_file, "CREATOR",std::string("adas preprocessing ").append(PREPROC_VERSION).c_str()," Program and version that created the FITS", &status);
	fits_write_key_str(fits_file, "FILENAME",fitsFileName.c_str(),"Filename", &status);
	int16_t filevers = 1;
	fits_write_key(fits_file, TSHORT, "FILEVERS",&filevers,"File version", &status);
	fits_write_key_str(fits_file, "ORIGIN_ID",database_params::orig_id.c_str(),"Date origin and context", &status);
	fits_write_key_str(fits_file, "OBS_DATE",filename_params::date.c_str(),"Date of the beginning of the observation night", &status);
	fits_write_key_str(fits_file, "OBS_MODE",filename_params::obs_mode.c_str(),"Observation mode", &status);
	fits_write_key_str(fits_file, "PROG_ID",database_params::prog_id.c_str(),"Observation program ID", &status);
	fits_write_key_str(fits_file, "SBL_ID",database_params::sbl_id.c_str(),"Observation scheduling block", &status);
	fits_write_key_str(fits_file, "RUN_ID",filename_params::run_id.c_str(),"Incremental array-wise file counter", &status);
	fits_write_key_str(fits_file, "DAQ_MODE",filename_params::mode_id.c_str(),"DAQ status identification", &status);
	fits_write_key_str(fits_file, "DAQ_ID",filename_params::daq_id.c_str(),"Telescope-wise DAQ incremental counter", &status);
	fits_write_key_str(fits_file, "DATAMODE","10","Run type", &status);
	fits_write_key_str(fits_file, "SUBMODE","01","Run subtype", &status);
	fits_write_key(fits_file, TSHORT, "NTEL",&database_params::ntel,"Number of active telescopes during observation", &status);
	fits_write_key_str(fits_file, "TIMESYS","TT","Time reference system", &status);
	int32_t MJDREFI = 58849;
	fits_write_key(fits_file, TINT, "MJDREFI",&MJDREFI,"MJD reference date", &status);
	float MJDREFF = 0.00080074;
	fits_write_key(fits_file, TFLOAT, "MJDREFF",&MJDREFF,"MJD reference fraction of day", &status);
	float TIMEOFFS = 0.0;
	fits_write_key(fits_file, TFLOAT, "TIMEOFFS",&MJDREFF,"Time offset", &status);
	fits_write_key_str(fits_file, "TIMEUNIT","s","Time unit", &status);
	fits_write_key_str(fits_file, "TSTART","X","First event time", &status);
	fits_write_key_str(fits_file, "TSTOP","X","Last event time", &status);
	fits_write_key_str(fits_file, "TELAPSE","X","Diff of start and end times", &status);
	fits_write_key_str(fits_file, "DATE-OBS","X","Start date of the first data event (UTC)", &status);
	fits_write_key_str(fits_file, "DATE-END","X","End date of the last data event (UTC)", &status);
    auto now = std::chrono::system_clock::now();
    std::time_t time_now = std::chrono::system_clock::to_time_t(now);
    std::tm tm_utc = *std::gmtime(&time_now);
    std::stringstream oss;
    oss << std::put_time(&tm_utc, "%Y-%m-%dT%H:%M:%S");
	fits_write_key_str(fits_file, "DATE",oss.str().c_str(),"File creation date (UTC)", &status);
	fits_write_key_str(fits_file, "OBJECT",filename_params::target_id.c_str(),"Name of observed object", &status);
	fits_write_key_str(fits_file, "RADECSYS","FK5","Coordinate reference frame", &status);
	fits_write_key_str(fits_file, "EQUINOX","2000.0","Coordinate epoch", &status);
	float RA_OBJ = 83.6331; // TODO check se questi paramtri da db o fissi
	fits_write_key(fits_file, TFLOAT, "RA_OBJ",&RA_OBJ,"[deg] R.A. Object", &status);
	float DEC_OBJ = 22.0145 ;
	fits_write_key(fits_file, TFLOAT, "DEC_OBJ",&DEC_OBJ,"[deg] Decl. Objec", &status);
	float RA_PNT = 84.133333 ;
	fits_write_key(fits_file, TFLOAT, "RA_PNT",&RA_PNT,"Pointed R.A. (J2000)", &status);
	float DEC_PNT = 22.014444;
	fits_write_key(fits_file, TFLOAT, "DEC_PNT",&DEC_PNT,"Pointed Dec. (J2000", &status);
	fits_write_key_str(fits_file, "CHECKSUM","g5XEj3XBg3XBg3XB","HDU checksum updated", &status);
	fits_write_key_str(fits_file, "DATASUM","0","Data unit checksum updated", &status);

	fits_create_tbl(fits_file, BINARY_TBL, 0, ncol, ttype, tform,
			tunit, extname, &status);
	if (status != 0) {
		fits_report_error(stderr, status); // print error message
		fits_delete_file(fits_file, &status);
		return -1;
	}


	status = 0;
    fits_movabs_hdu(fits_file, 2, nullptr, &status);

	if (status != 0) {
		fits_report_error(stderr, status); // print error message
		fits_delete_file(fits_file, &status);
		return -1;
	}
	return 1;
}

static int addHK101StartTimesToFits(Packets::AstriMaGeneric *packet,
		fitsfile *fitsfile) {
	int status = 0;
	fits_movabs_hdu(fitsfile, 1, nullptr, &status);

	int year = packet->operator []("year").value();
	int month = packet->operator []("month").value();
	int day = packet->operator []("day").value();
	int hours = packet->operator []("hours").value();
	int minutes = packet->operator []("minutes").value();
	int seconds = packet->operator []("seconds").value();

	double d1 = 0, d2 = 0;
	iauDtf2d("UTC", year, month, day, hours, minutes, seconds, &d1, &d2);
	FitsElements::HK101::start_time = d1+d2;

	fits_update_key_str(fitsfile, "TSTART", std::to_string(FitsElements::HK101::start_time).c_str(), "First event time", &status);

	std::stringstream ss;
	ss << std::setfill('0') << std::setw(4) << year;
	ss << "-";
	ss << std::setfill('0') << std::setw(2) << month;
	ss << "-";
	ss << std::setfill('0') << std::setw(2) << day;
	ss << "T";
	ss << std::setfill('0') << std::setw(2) << hours;
	ss << ":";
	ss << std::setfill('0') << std::setw(2) << minutes;
	ss << ":";
	ss << std::setfill('0') << std::setw(2) << seconds;
	std::string time_utc = ss.str();
	fits_update_key_str(fitsfile, "DATE-OBS", time_utc.c_str(),
			"Start date of the first data event (UTC)", &status);
	fits_movabs_hdu(fitsfile, 2, nullptr, &status);
	return 1;
}
static int addHK101EndTimesToFits(fitsfile *fitsfile) {
	using namespace FitsElements::HK101::last_event_time;
	int status = 0;
	fits_movabs_hdu(fitsfile, 1, nullptr, &status);

	double d1 = 0, d2 = 0;
	iauDtf2d("UTC", year, month, day, hours, minutes, seconds, &d1, &d2);
	FitsElements::HK101::stop_time = d1 + d2;
	fits_update_key_str(fitsfile, "TSTOP", std::to_string(FitsElements::HK101::stop_time).c_str(), "Last event time", &status);

	double telapse = FitsElements::HK101::stop_time - FitsElements::HK101::start_time;
	fits_update_key_str(fitsfile, "TELAPSE", std::to_string(telapse).c_str(), "Diff of start and end times",
			&status);

	fits_update_key_str(fitsfile, "DATE-END", FitsElements::HK101::time_utc.c_str(),
			"End date of the last data event (UTC)", &status);
	fits_movabs_hdu(fitsfile, 2, nullptr, &status);
	return 1;
}


static int addHK101EventPacketToFits(Packets::AstriMaGeneric *packet, fitsfile *fitsfile, int nrow) {
	using namespace FitsElements::HK101;
	int count = 0;
	int res = 0;

	bool vt = packet->operator []("vt").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &vt, &res);
	last_event_time::year = packet->operator []("year").value();
	last_event_time::month = packet->operator []("month").value();
	last_event_time::day = packet->operator []("day").value();
	last_event_time::hours = packet->operator []("hours").value();
	last_event_time::minutes = packet->operator []("minutes").value();
	last_event_time::seconds = packet->operator []("seconds").value();
	std::stringstream ss;
	ss << std::setfill('0') << std::setw(4) << last_event_time::year ;
	ss << "-";
	ss << std::setfill('0') << std::setw(2) << last_event_time::month ;
	ss << "-";
	ss << std::setfill('0') << std::setw(2) << last_event_time::day;
	ss << "T";
	ss << std::setfill('0') << std::setw(2) << last_event_time::hours;
	ss << ":";
	ss << std::setfill('0') << std::setw(2) << last_event_time::minutes;
	ss << ":";
	ss << std::setfill('0') << std::setw(2) << last_event_time::seconds;
	FitsElements::HK101::time_utc = ss.str();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &FitsElements::HK101::time_utc, &res);
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &last_event_time::seconds, &res);
	uint32_t time_tag_ns = packet->operator []("Time Tag Nanoseconds").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &time_tag_ns, &res);
	uint16_t ssc = packet->operator []("Source Sequence Counter SSC").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &ssc, &res);
	uint16_t packet_len = packet->operator []("Packet Length").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &packet_len, &res);
	uint32_t evt_counter = packet->operator []("Event Counter").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &evt_counter, &res);
	bool lid = packet->operator []("lid").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &lid, &res);
	bool fib_st = packet->operator []("fib st").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &fib_st, &res);
	bool fib_cont = packet->operator []("fib cont").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &fib_cont, &res);
	bool fib_puls = packet->operator []("fib puls").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &fib_puls, &res);
	int16_t rgb_cont = packet->operator []("rgb cont").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &rgb_cont, &res);
	int16_t rgb_puls = packet->operator []("rgb puls").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &rgb_puls, &res);
	int16_t cam_tmp1 = packet->operator []("Camera Temperature 1").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &cam_tmp1, &res);
	int16_t cam_tmp2 = packet->operator []("Camera Temperature 2").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &cam_tmp2, &res);
	int16_t cam_tmp3 = packet->operator []("Camera Temperature 3").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &cam_tmp3, &res);
	int16_t cam_tmp4 = packet->operator []("Camera Temperature 4").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &cam_tmp4, &res);
	int16_t cam_hum1 = packet->operator []("Camera Humidity 1").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &cam_hum1, &res);
	int16_t cam_hum2 = packet->operator []("Camera Humidity 2").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &cam_hum2, &res);
	int16_t cam_hum3 = packet->operator []("Camera Humidity 3").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &cam_hum3, &res);
	int16_t cam_hum4 = packet->operator []("Camera Humidity 4").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &cam_hum4, &res);
	int16_t bl_frq = packet->operator []("Blu laser frequency").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &bl_frq, &res);
	int16_t bl_dur = packet->operator []("Blu laser duration").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &bl_dur, &res);
	int16_t bl_v = packet->operator []("Blu laser voltage").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &bl_v, &res);
	int16_t gl_frq = packet->operator []("Green laser frequency").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &gl_frq, &res);
	int16_t gl_dur = packet->operator []("Green laser duration").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &gl_dur, &res);
	int16_t gl_v = packet->operator []("Green laser voltage").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &gl_v, &res);
	int16_t ld_tmp = packet->operator []("Laser diodes temperature").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &ld_tmp, &res);
	int16_t cam_rtm = packet->operator []("Camera rate meter").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &cam_rtm, &res);
	int16_t tcam_rtm = packet->operator []("Total Camera rate meter").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &tcam_rtm, &res);
	int16_t dead_time = packet->operator []("Dead time").value();
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, 1, &dead_time, &res);
	bool pdm_vals[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "pdm val_#" << i;
		pdm_vals[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &pdm_vals, &res);
	uint16_t pdm_tmp1[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "SiPM temperature 1_#" << i;
		pdm_tmp1[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &pdm_tmp1, &res);
	uint16_t pdm_tmp2[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "SiPM temperature 2_#" << i;
		pdm_tmp2[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &pdm_tmp2, &res);
	uint16_t pdm_tmp3[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "SiPM temperature 3_#" << i;
		pdm_tmp3[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &pdm_tmp3, &res);
	uint16_t pdm_cbt[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "CITIROC board temperature_#" << i;
		pdm_cbt[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &pdm_cbt, &res);
	uint16_t pdm_cbv1[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "CITIROC board voltage 1_#" << i;
		pdm_cbv1[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &pdm_cbv1, &res);
	uint16_t pdm_cbv2[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "CITIROC board voltage 2_#" << i;
		pdm_cbv2[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &pdm_cbv2, &res);
	uint16_t pdm_cbc1[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "CITIROC board current 1_#" << i;
		pdm_cbc1[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &pdm_cbc1, &res);
	uint16_t pdm_cbc2[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "CITIROC board current 2_#" << i;
		pdm_cbc2[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &pdm_cbc2, &res);
	uint16_t pdm_cbc3[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "CITIROC board current 3_#" << i;
		pdm_cbc3[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &pdm_cbc3, &res);
	uint16_t fpga_tmp[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "FPGA temperature_#" << i;
		fpga_tmp[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &fpga_tmp, &res);
	uint16_t fpga_v[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "FPGA voltage_#" << i;
		fpga_v[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &fpga_v, &res);
	uint16_t sipm_hv[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "SiPM High Voltage_#" << i;
		sipm_hv[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &sipm_hv, &res);
	uint32_t sipm_c[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "SiPM_current_#" << i;
		sipm_c[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &sipm_c, &res);
	uint16_t cam_rt[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "Camera Rate_#" << i;
		cam_rt[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &cam_rt, &res);
	uint16_t pdm_trt[num_pdm];
	for (int i = 0; i < num_pdm; i++) {
		std::stringstream ss;
		ss << "PDM Trigger rate_#" << i;
		pdm_trt[i] = packet->operator [](ss.str()).value();
	}
	fits_write_col(fitsfile,
			std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
			nrow, 1, num_pdm, &pdm_trt, &res);
	uint32_t pdm_pixel_rate[num_pdm][num_pdm_pixel_rate];
	for (int i = 0; i < num_pdm; i++) {
		for (int j = 0; j < num_pdm_pixel_rate; j++) {
			std::stringstream ss;
			ss << "PDM Pixel Rate_#" << j << "_#" << i;
			pdm_pixel_rate[i][j] = packet->operator [](ss.str()).value();
		}
		fits_write_col(fitsfile,
				std::get<3>(FitsElements::HK101::events_hdu[count++]), count + 1,
				nrow, 1, num_pdm_pixel_rate, &pdm_pixel_rate[i], &res);
	}

	return 1;
}

};
