#include <AstriFitsProcessor.h>
#include <FitsHelperC11.h>
#include <FitsHelperC14.h>
#include <FitsHelperHK101.h>
#include <FitsHelperVAR102.h>
#include <FitsHelperVAR103.h>
#include <FitsHelperS22.h>
#include <regex>
#include <constants.h>
#include <filesystem>

using namespace inaf::oasbo;

void AstriFitsProcessor::printLog(std::ostream &os, std::string message) {
	time_t now = time(nullptr);
	os << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[Fits Processor]" << "\t" << message << std::endl;
}

AstriFitsProcessor::AstriFitsProcessor(std::string file_to_process) {
	this->file_to_process = file_to_process;
	this->destinationPath = std::string("./"); // current dir
	this->receiver =
			Receivers::FileReceiverBuilder().setSource(file_to_process)->setRate(
					0)->getReceiver();

	// init packet with gen structure
	std::string pathGen(
			std::string(ASTRI_PACKETS_PATH).append(
					"/AstriMA/AstriMAGeneric.json"));
	this->structureGen = new Packets::PacketStructureJson(pathGen);
	this->packet = new Packets::AstriMaGeneric(*structureGen);

	// prepare other packet structures
	std::string pathC11(
			std::string(ASTRI_PACKETS_PATH).append("/AstriMA/11.json"));
	this->structureC11 = new Packets::PacketStructureJson(pathC11);

	std::string pathC14(
			std::string(ASTRI_PACKETS_PATH).append("/AstriMA/14.json"));
	this->structureC14 = new Packets::PacketStructureJson(pathC14);

	std::string pathS22(
			std::string(ASTRI_PACKETS_PATH).append("/AstriMA/22.json"));
	this->structureS22 = new Packets::PacketStructureJson(pathS22);

	std::string pathVHG(
			std::string(ASTRI_PACKETS_PATH).append("/AstriMA/102.json"));
	this->structureVHG = new Packets::PacketStructureJson(pathVHG);

	std::string pathVLG(
			std::string(ASTRI_PACKETS_PATH).append("/AstriMA/103.json"));
	this->structureVLG = new Packets::PacketStructureJson(pathVLG);

	std::string pathHK(
			std::string(ASTRI_PACKETS_PATH).append("/AstriMA/101.json"));
	this->structureHK = new Packets::PacketStructureJson(pathHK);

	FitsFunctions::initC11FitsSchema();
	FitsFunctions::initC14FitsSchema();
	FitsFunctions::initHK101FitsSchema();
	FitsFunctions::initVAR102FitsSchema();
	FitsFunctions::initVAR103FitsSchema();
	FitsFunctions::initS22FitsSchema();
}

void AstriFitsProcessor::setFileToProcess(std::string file_to_process) {
	this->receiver->setHost(file_to_process);
	this->file_to_process = file_to_process;
}

std::string AstriFitsProcessor::getFileToProcess() {
	return this->file_to_process;
}

int AstriFitsProcessor::startProcessing() {
	if (!this->isFileNameConform()) {
		printLog(std::cerr,
				" Error: File name is not conform to naming convention:\n\tFileName:"
						+ this->file_to_process);
		return -1;
	}

	using namespace FitsElements;
	if (this->receiver->connectToClient() < 0) {
		printLog(std::cerr, "Error, cannot open file: " + receiver->getHost());
		return -1;
	}

	// C11
	fitsfile *fptr_c11;
	std::string fits11Filename;
	bool firstC11 = true;
	int nrow11 = 0;

	// C14
	fitsfile *fptr_c14;
	std::string fits14Filename;
	bool firstC14 = true;
	int nrow14 = 0;

	// HK101
	fitsfile *fptr_hk101;
	std::string fits101Filename;
	bool firstHK101 = true;
	int nrow101 = 0;

	// VAR102
	fitsfile *fptr_var102;
	std::string fits102Filename;
	bool firstVAR102 = true;
	int nrow102 = 0;

	// VAR103
	fitsfile *fptr_var103;
	std::string fits103Filename;
	bool firstVAR103 = true;
	int nrow103 = 0;

	//S22
	fitsfile *fptr_s22;
	std::string fits22Filename;
	bool firstS22 = true;
	int nrow22 = 0;

	while (true) { // go ahead until no packet to read or error occurred
		int rcv = this->receiver->receiveFromClient(*this->packet);
		if (rcv > 0) { // packet read
			// C11
			if (packet->getType() == 1 && packet->getSubType() == 1) {
				if (firstC11) {
					fits11Filename = computeFitsFileName("0101");
					int status = 0;
					fits_create_file(&fptr_c11,
							(this->destinationPath + fits11Filename).c_str(),
							&status);
					if (status != 0) {
						fits_report_error(stderr, status); // print error message
						return -1;
					}
					status = 0;
					int done = FitsFunctions::initFitsFileC11(fptr_c11,
							fits11Filename, this->file_to_process);
					if (done < 1) { // error initiating the file
						printLog(std::cerr,
								"Error error initiating the file "
										+ fits11Filename);
						return -1; // exit
					}
					FitsFunctions::addC11StartTimesToFits(this->packet,
							fptr_c11);
					firstC11 = false;
				}
				nrow11++;
				this->packet->updatePacketStructure(*this->structureC11);
				FitsFunctions::addC11EventPacketToFits(this->packet, fptr_c11,
						nrow11);
				this->packet->updatePacketStructure(*this->structureGen);
				continue;
			}
			// C14
			if (packet->getType() == 1 && packet->getSubType() == 4) {
				if (firstC14) {
					fits14Filename = computeFitsFileName("0104");
					int status = 0;
					fits_create_file(&fptr_c14,
							(this->destinationPath + fits14Filename).c_str(),
							&status);
					if (status != 0) {
						fits_report_error(stderr, status); // print error message
						return -1;
					}
					status = 0;
					int done = FitsFunctions::initFitsFileC14(fptr_c14,
							fits14Filename, this->file_to_process);
					if (done < 1) { // error initiating the file
						printLog(std::cerr,
								"Error error initiating the file "
										+ fits14Filename);
						return -1; // exit
					}
					FitsFunctions::addC14StartTimesToFits(this->packet,
							fptr_c14);
					firstC14 = false;
				}
				nrow14++;
				this->packet->updatePacketStructure(*this->structureC14);
				FitsFunctions::addC14EventPacketToFits(this->packet, fptr_c14,
						nrow14);
				this->packet->updatePacketStructure(*this->structureGen);
				continue;
			}
			// HK101
			if (packet->getType() == 10 && packet->getSubType() == 1) {
				if (firstHK101) {
					fits101Filename = computeFitsFileName("1001");
					int status = 0;
					fits_create_file(&fptr_hk101,
							(this->destinationPath + fits101Filename).c_str(),
							&status);
					if (status != 0) {
						fits_report_error(stderr, status); // print error message
						return -1;
					}
					status = 0;
					int done = FitsFunctions::initFitsFileHK101(fptr_hk101,
							fits101Filename, this->file_to_process);
					if (done < 1) { // error initiating the file
						printLog(std::cerr,
								"Error error initiating the file "
										+ fits101Filename);
						return -1; // exit
					}
					FitsFunctions::addHK101StartTimesToFits(this->packet,
							fptr_hk101);
					firstHK101 = false;
				}
				nrow101++;
				this->packet->updatePacketStructure(*this->structureHK);
				FitsFunctions::addHK101EventPacketToFits(this->packet,
						fptr_hk101, nrow101);
				this->packet->updatePacketStructure(*this->structureGen);
				continue;
			}
			// VAR102
			if (packet->getType() == 10 && packet->getSubType() == 2) {
				if (firstVAR102) {
					fits102Filename = computeFitsFileName("1002");
					int status = 0;
					fits_create_file(&fptr_var102,
							(this->destinationPath + fits102Filename).c_str(),
							&status);
					if (status != 0) {
						fits_report_error(stderr, status); // print error message
						return -1;
					}
					status = 0;
					int done = FitsFunctions::initFitsFileVAR102(fptr_var102,
							fits102Filename, this->file_to_process);
					if (done < 1) { // error initiating the file
						printLog(std::cerr,
								"Error error initiating the file "
										+ fits102Filename);
						return -1; // exit
					}
					FitsFunctions::addVAR102StartTimesToFits(this->packet,
							fptr_var102);
					firstVAR102 = false;
				}
				nrow102++;
				this->packet->updatePacketStructure(*this->structureVHG);
				FitsFunctions::addVAR102EventPacketToFits(this->packet,
						fptr_var102, nrow102);
				this->packet->updatePacketStructure(*this->structureGen);
				continue;
			}
			// VAR103
			if (packet->getType() == 10 && packet->getSubType() == 3) {
				if (firstVAR103) {
					fits103Filename = computeFitsFileName("1003");
					int status = 0;
					fits_create_file(&fptr_var103,
							(this->destinationPath + fits103Filename).c_str(),
							&status);
					if (status != 0) {
						fits_report_error(stderr, status); // print error message
						return -1;
					}
					status = 0;
					int done = FitsFunctions::initFitsFileVAR103(fptr_var103,
							fits103Filename, this->file_to_process);
					if (done < 1) { // error initiating the file
						printLog(std::cerr,
								"Error error initiating the file "
										+ fits103Filename);
						return -1; // exit
					}
					FitsFunctions::addVAR103StartTimesToFits(this->packet,
							fptr_var103);
					firstVAR103 = false;
				}
				nrow103++;
				this->packet->updatePacketStructure(*this->structureVLG);
				FitsFunctions::addVAR103EventPacketToFits(this->packet,
						fptr_var103, nrow103);
				this->packet->updatePacketStructure(*this->structureGen);
				continue;
			}
			//S22
			if (packet->getType() == 2 && packet->getSubType() == 2) {
				if (firstS22) {
					fits22Filename = computeFitsFileName("0202");
					int status = 0;
					fits_create_file(&fptr_s22,
							(this->destinationPath + fits22Filename).c_str(),
							&status);
					if (status != 0) {
						fits_report_error(stderr, status); // print error message
						return -1;
					}
					status = 0;
					int done = FitsFunctions::initFitsFileS22(fptr_s22,
							fits22Filename, this->file_to_process);
					if (done < 1) { // error initiating the file
						printLog(std::cerr,
								"Error error initiating the file "
										+ fits22Filename);
						return -1; // exit
					}
					FitsFunctions::addS22StartTimesToFits(this->packet,
							fptr_s22);
					firstS22 = false;
				}
				nrow22++;
				this->packet->updatePacketStructure(*this->structureS22);
				FitsFunctions::addS22EventPacketToFits(this->packet, fptr_s22,
						nrow22);
				this->packet->updatePacketStructure(*this->structureGen);
				continue;
			}
			// other packets
		}
		if (rcv < 0) { // error reading packet
			printLog(std::cerr, "Error reading packet");
			continue; // go to next packet
		}
		if (rcv == 0) { // no more packets to read
			int status = 0;
			// C11
			if (!firstC11) { //at least one packet received -> file created
				FitsFunctions::addC11EndTimesToFits(fptr_c11);
				fits_close_file(fptr_c11, &status);
				printLog(std::cout, "File C11 completed: " + fits11Filename);
			}
			// C14
			if (!firstC14) { //at least one packet received -> file created
				FitsFunctions::addC14EndTimesToFits(fptr_c14);
				fits_close_file(fptr_c14, &status);
				printLog(std::cout, "File C14 completed: " + fits14Filename);
			}
			// HK101
			if (!firstHK101) { //at least one packet received -> file created
				FitsFunctions::addHK101EndTimesToFits(fptr_hk101);
				fits_close_file(fptr_hk101, &status);
				printLog(std::cout,
						"File HK101 completed: " + fits101Filename);
			}
			// VAR102
			if (!firstVAR102) { //at least one packet received -> file created
				FitsFunctions::addVAR102EndTimesToFits(fptr_var102);
				fits_close_file(fptr_var102, &status);
				printLog(std::cout,
						"File VAR102 completed: " + fits102Filename);
			}
			// VAR103
			if (!firstVAR103) { //at least one packet received -> file created
				FitsFunctions::addVAR103EndTimesToFits(fptr_var103);
				fits_close_file(fptr_var103, &status);
				printLog(std::cout,
						"File VAR103 completed: " + fits103Filename);
			}
			// S22
			if (!firstS22) { //at least one packet received -> file created
				FitsFunctions::addS22EndTimesToFits(fptr_s22);
				fits_close_file(fptr_s22, &status);
				printLog(std::cout, "File S22 completed: " + fits22Filename);
			}
			break; // exit
		}
	}

	return 1;
}

bool AstriFitsProcessor::isFileNameConform() {
	std::regex pattern(RAW_FILE_NAMING_CONV);
	std::filesystem::path filePath(this->file_to_process);
	std::string fileName(filePath.filename());
	if (std::regex_match(fileName, pattern)) {
		return true;
	}
	return false;
}

std::string AstriFitsProcessor::computeFitsFileName(std::string type_subtype) {
	std::filesystem::path rawfilePath(this->file_to_process);
	std::string rawfileName = rawfilePath.filename().string();
	std::string fitsFileName(
			rawfileName.substr(0, rawfileName.length() - 4).append("_").append(type_subtype).append(".lv0.fits"));
	return fitsFileName;
}

int AstriFitsProcessor::setDestinationPath(std::string destPath) {
	if (!destPath.back() != '/') {
		destPath += '/';
	}
	namespace fs = std::filesystem;
	fs::path destination(destPath);
	if (!fs::exists(destination)) {
		if (!fs::create_directories(destination)) {
			printLog(std::cerr,
					"Error: Unable to create directory " + destPath + "; using current directory");
			return -1;
		}
	}
	this->destinationPath = destPath;
	return 1;
}

std::string AstriFitsProcessor::getDestinationPath() {
	return this->destinationPath;
}
